#pragma once

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

// Ett linjesegment mellan punkten (x1, y1) och (x2, y2).
struct line {
    // Linjens startpunkt.
    int x1;
    int y1;

    // Linjens slutpunkt.
    int x2;
    int y2;
};


// En bild. Bilden består av ett antal linjesegment.
struct picture {
    // Bredd och höjd på bilden.
    int width;
    int height;

    // Array av linjesegment som bilden består av.
    struct line *lines;

    // Antal linjesegment i arrayen "lines".
    int line_count;
};


/**
 * Nedan finns funktioner som hanterar bilder.
 */


// Läs in en bild från en fil.
//
// Denna funktion används för att skapa och läsa in "picture"-objekt.
// Funktionerna "fopen", "fscanf" och "fclose" används för att öppna,
// läsa från, och stänga filer. Dessa funktioner är trådsäkra på det
// sätt de används här.
struct picture *picture_load(const char *file) {
    // Öppna filen.
    FILE *f = fopen(file, "r");
    if (f == NULL)
        return NULL;

    // Skapa en picture-instans och läs storlek.
    struct picture *result = malloc(sizeof(struct picture));
    fscanf(f, "%d %d\n", &result->width, &result->height);
    fscanf(f, "%d\n", &result->line_count);

    // Allokera minne för alla linjesegment, och läs dem från filen.
    result->lines = malloc(result->line_count * sizeof(struct line));
    for (int i = 0; i < result->line_count; i++) {
        fscanf(f, "%d %d %d %d\n",
            &result->lines[i].x1,
            &result->lines[i].y1,
            &result->lines[i].x2,
            &result->lines[i].y2);
    }

    fclose(f);

    return result;
}


// Förstör en bild.
void picture_free(struct picture *p) {
    free(p->lines);
    free(p);
}


// Kontrollera ifall punkten (x, y) ligger på linjesegmentet "l".
//
// Denna funktion används av resten av koden för att bestämma ifall
// en viss "pixel" ska fyllas i eller inte. Du behöver inte fördjupa
// dig i exakt hur koden fungerar. Det är dock intressant att observera
// att funktionen inte använder data utöver sina parametrar.
bool point_in_line(int x, int y, struct line l) {
    double dirx = l.x2 - l.x1;
    double diry = l.y2 - l.y1;
    double ptx = x - l.x1;
    double pty = y - l.y1;

    double length = dirx*dirx + diry*diry;
    if (length < 0.01)
        length = 1;
    double dot = (dirx*ptx + diry*pty) / length;
    if (dot < 0.0 || dot > 1.0)
        return false;

    double distx = dot*dirx - ptx;
    double disty = dot*diry - pty;
    return (distx*distx + disty*disty) < 0.4;
}


// Generera pixlar för raden med koordinat "y" i bilden "p".
char *picture_compute_row(const struct picture *p, int y) {
    char *line = malloc(p->width + 1);
    for (int x = 0; x < p->width; x++) {

        // Se ifall något linjesegment ligger på den här pixeln.
        bool filled = false;
        for (int i = 0; i < p->line_count; i++) {
            if (point_in_line(x, y, p->lines[i])) {
                filled = true;
            }
        }

        // Fyll i strängen.
        if (filled)
            line[x] = 'x';
        else
            line[x] = ' ';
    }
    // Se till att strängen slutar med tecknet noll.
    line[p->width] = '\0';
    return line;
}
